#  XTrkCad - Model Railroad CAD
#  Copyright (C) 2016 Martin Fischer
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# Create the bitmap used for layer buttons, requires ImageMagick and the GohuFont 14pt

$font = "GohuFont"
$buttons = 99

for($i=1; $i -le $buttons; $i++ )
{
    convert -size 14x16  -gravity center -pointsize 14  -font $font label:$i l$i.xbm
}