/************************************************************************************
LINUX Help File colour and fonts are done here. Halibut controls margins, line spacing and indents.
*************************************************************************************/

BODY {
	background: white;
	font-family: "Trebuchet MS", "Lucida Grande", "Lucida Sans Unicode", "Lucida Sans", Tahoma, sans-serif;
	font-size: 100%;
	color: #333;
	font-weight: normal;
	line-height: 1.4em;
}

A:link {
	text-decoration: none;
	color: #7DA1BF;
	font-weight: bold;
}          /* unvisited link */

A:visited {
	text-decoration: none;
	color: #1d4e89;
	font-weight: bold;
}        /* visited links */

A:active { 
	text-decoration: none;
	font-weight: bold;	
	color: red;
	}        /* active links */

a:hover {
	text-decoration: underline;
	color: #F69256;
	}

H1 {
	text-align: left;
	font-size: large;
	font-style: italic;
	font-weight: bold;
	padding: 3px 6px;
}

H2 {
	text-align: left;
	font-size: medium;
	font-style: italic;
	font-weight: bold;
	padding: 3px 6px;
}

H3 {
	text-align: left;
	font-size: small;
	font-style: italic;
	font-weight: bold;
	padding: 3px 6px;
}
