#ifndef _BEZCTX_H
#define _BEZCTX_H
#include "bezctx_intf.h"

struct _bezctx {
    void (*moveto)(bezctx *bc, double x, double y, int is_open);
    void (*lineto)(bezctx *bc, double x, double y);
    void (*quadto)(bezctx *bc, double x1, double y1, double x2, double y2);
    void (*curveto)(bezctx *bc, double x1, double y1, double x2, double y2,
		    double x3, double y3);
    void (*mark_knot)(bezctx *bc, int knot_idx);
};
#endif
