/* Some colors */
//#define wDrawColorWhite		(0)
//#define wDrawColorBlack		(1)
//#define wDrawColorRed		(2)
//#define wDrawColorGreen		(3)
//#define wDrawColorBlue		(4)
//#define wDrawColorYellow	(5)
//#define wDrawColorPurple	(6)
//#define wDrawColorAqua		(7)
//#define wDrawColorDkRed		(8)
//#define wDrawColorDkGreen	(9)
//#define wDrawColorDkBlue	(10)
//#define wDrawColorDkYellow	(11)
//#define wDrawColorDkPurple	(12)
//#define wDrawColorDkAqua	(13)
//#define wDrawColorRoyalBlue	(14)
//#define wDrawColorDeepSkyBlue	(15)
//#define wDrawColorLightSkyBlue	(16)
//#define wDrawColorSteelBlue	(17)
//#define wDrawColorPowderBlue	(18)
//#define wDrawColorAquamarine	(19)
//#define wDrawColorSeaGreen	(20)
//#define wDrawColorPaleGreen	(21)
//#define wDrawColorLawnGreen	(22)
//#define wDrawColorLimeGreen	(23)
//#define wDrawColorForestGreen	(24)
//#define wDrawColorGold		(25)
//#define wDrawColorRosyBrown	(26)
//#define wDrawColorSaddleBrown	(27)
//#define wDrawColorBeige		(28)
//#define wDrawColorTan		(29)
//#define wDrawColorChocolate	(30)
//#define wDrawColorBrown		(31)
//#define wDrawColorOrange	(32)
//#define wDrawColorCoral		(33)
//#define wDrawColorTomato	(34)
//#define wDrawColorHotPink	(35)
//#define wDrawColorPink		(36)
//#define wDrawColorMaroon	(37)
//#define wDrawColorViolet	(38)
//#define wDrawColorPurple2	(39)

