/*
 * sbcsdat.h - header file for SBCS data structures.
 *
 * Generated by sbcsgen.pl from sbcs.dat.
 * You should edit those files rather than editing this one.
 */

#ifndef charset_sbcsdat_h
#define charset_sbcsdat_h

#include "charset.h"
#include "internal.h"

extern const sbcs_data sbcsdata_CS_ASCII;
extern const sbcs_data sbcsdata_CS_BS4730;
extern const sbcs_data sbcsdata_CS_DEC_GRAPHICS;
extern const sbcs_data sbcsdata_CS_ISO8859_1;
extern const sbcs_data sbcsdata_CS_ISO8859_2;
extern const sbcs_data sbcsdata_CS_ISO8859_3;
extern const sbcs_data sbcsdata_CS_ISO8859_4;
extern const sbcs_data sbcsdata_CS_ISO8859_5;
extern const sbcs_data sbcsdata_CS_ISO8859_6;
extern const sbcs_data sbcsdata_CS_ISO8859_7;
extern const sbcs_data sbcsdata_CS_ISO8859_8;
extern const sbcs_data sbcsdata_CS_ISO8859_9;
extern const sbcs_data sbcsdata_CS_ISO8859_10;
extern const sbcs_data sbcsdata_CS_ISO8859_11;
extern const sbcs_data sbcsdata_CS_ISO8859_13;
extern const sbcs_data sbcsdata_CS_ISO8859_14;
extern const sbcs_data sbcsdata_CS_ISO8859_15;
extern const sbcs_data sbcsdata_CS_ISO8859_16;
extern const sbcs_data sbcsdata_CS_ISO8859_1_X11;
extern const sbcs_data sbcsdata_CS_CP437;
extern const sbcs_data sbcsdata_CS_CP850;
extern const sbcs_data sbcsdata_CS_CP866;
extern const sbcs_data sbcsdata_CS_CP1250;
extern const sbcs_data sbcsdata_CS_CP1251;
extern const sbcs_data sbcsdata_CS_CP1252;
extern const sbcs_data sbcsdata_CS_CP1253;
extern const sbcs_data sbcsdata_CS_CP1254;
extern const sbcs_data sbcsdata_CS_CP1255;
extern const sbcs_data sbcsdata_CS_CP1256;
extern const sbcs_data sbcsdata_CS_CP1257;
extern const sbcs_data sbcsdata_CS_CP1258;
extern const sbcs_data sbcsdata_CS_KOI8_R;
extern const sbcs_data sbcsdata_CS_KOI8_U;
extern const sbcs_data sbcsdata_CS_KOI8_RU;
extern const sbcs_data sbcsdata_CS_JISX0201;
extern const sbcs_data sbcsdata_CS_MAC_ROMAN;
extern const sbcs_data sbcsdata_CS_MAC_TURKISH;
extern const sbcs_data sbcsdata_CS_MAC_CROATIAN;
extern const sbcs_data sbcsdata_CS_MAC_ICELAND;
extern const sbcs_data sbcsdata_CS_MAC_ROMANIAN;
extern const sbcs_data sbcsdata_CS_MAC_GREEK;
extern const sbcs_data sbcsdata_CS_MAC_CYRILLIC;
extern const sbcs_data sbcsdata_CS_MAC_THAI;
extern const sbcs_data sbcsdata_CS_MAC_CENTEURO;
extern const sbcs_data sbcsdata_CS_MAC_SYMBOL;
extern const sbcs_data sbcsdata_CS_MAC_DINGBATS;
extern const sbcs_data sbcsdata_CS_MAC_ROMAN_OLD;
extern const sbcs_data sbcsdata_CS_MAC_CROATIAN_OLD;
extern const sbcs_data sbcsdata_CS_MAC_ICELAND_OLD;
extern const sbcs_data sbcsdata_CS_MAC_ROMANIAN_OLD;
extern const sbcs_data sbcsdata_CS_MAC_GREEK_OLD;
extern const sbcs_data sbcsdata_CS_MAC_CYRILLIC_OLD;
extern const sbcs_data sbcsdata_CS_MAC_UKRAINE;
extern const sbcs_data sbcsdata_CS_MAC_VT100;
extern const sbcs_data sbcsdata_CS_MAC_VT100_OLD;
extern const sbcs_data sbcsdata_CS_VISCII;
extern const sbcs_data sbcsdata_CS_HP_ROMAN8;
extern const sbcs_data sbcsdata_CS_DEC_MCS;
extern const sbcs_data sbcsdata_CS_PDF;
extern const sbcs_data sbcsdata_CS_PSSTD;

#endif /* charset_sbcsdat_h */
