/** \file version.h
 *
 */

/*  XTrkCad - Model Railroad CAD
 *  Copyright (C) 2005 Dave Bullis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef HAVE_VERSION_H
#define HAVE_VERSION_H
#ifdef XTRKCAD_CMAKE_BUILD

#include "xtrkcad-config.h"

#define VERSION XTRKCAD_VERSION
#define PARAMVERSION XTRKCAD_PARAMVERSION
#define PARAMVERSIONVERSION XTRKCAD_PARAMVERSIONVERSION
#define MINPARAMVERSION XTRKCAD_MINPARAMVERSION

#else

#define VERSION "4.1.0b1"
#define PARAMVERSION (10)
#define PARAMVERSIONVERSION "3.0.0"
#define MINPARAMVERSION (1)

#endif
#endif //HAVE_VERSION_H
