#define IDD_DLG1 1000
#define IDC_IMG1 1001
#define IDC_STC1 1002
#define IDC_STC2 1003
#define IDC_IMG2 1004
#define IDI_LOGO 1000
IDD_DLG1 DIALOGEX 6,6,194,102
CAPTION "IDD_DLG"
FONT 8,"MS Sans Serif",0,0
STYLE 0x10CF0000
BEGIN
  CONTROL "",IDC_IMG1,"Static",0x50000203,196,99,92,26
  CONTROL "Testapplication",IDC_STC1,"Static",0x50000201,0,57,194,9
  CONTROL "Application is starting",IDC_STC2,"Static",0x50000000,-2,68,199,9
  CONTROL "",IDC_IMG2,"Static",0x50000203,0,0,226,57
END
IDI_LOGO IMAGE DISCARDABLE "logo.bmp"
