#include <windows.h>
#include "w.h"
#include "mswlib.h"
MSWAPPICON ICON wtest.ico
Generic MENU
BEGIN
	POPUP "&Help"
	BEGIN
		MENUITEM "About Generic...", IDM_ABOUT
	END
END

AboutBox DIALOG 22, 17, 144, 75
STYLE DS_MODALFRAME | WS_CAPTION | WS_SYSMENU
CAPTION "About Generic"
BEGIN
	CTEXT "WLIB Windows", -1, 0, 5, 144, 8
	CTEXT "Test Application", -1, 0, 14, 144, 8
	CTEXT "Version 0.1", -1, 0, 34, 144, 8
	DEFPUSHBUTTON "OK" IDOK, 53, 59, 32, 14, WS_GROUP
END
#include "mswlib.rc"
