msw-accel ACCELERATORS
BEGIN
	VK_F1, IDM_DOHELP, VIRTKEY
END

MswAbortDlg DIALOG 20, 20, 90, 64
STYLE DS_MODALFRAME | WS_CAPTION | WS_SYSMENU
CAPTION "Printing"
BEGIN
	DEFPUSHBUTTON "Cancel", IDCANCEL, 29, 44, 32, 14, WS_GROUP
	CTEXT	"Now printing", -1, 0, 8, 90, 8
	CTEXT	"Page", IDM_PRINTPAGE, 0, 18, 90, 8
END
